import eboot
import movie_bin_tool
import shared_bin_tool
import struct
import subprocess


#Constants for string insertion
INPUT_COL = 2
PTR_COL = 5
s1 = " .,'!?abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890~-&>():;/%"
s2 = "　．，’！？ａｂｃｄｅｆｇｈｉｊｋｌｍｎｏｐｑｒｓｔｕｖｗｘｙｚＡＢＣＤＥＦＧＨＩＪＫＬＭＮＯＰＱＲＳＴＵＶＷＸＹＺ１２３４５６７８９０～‐＆＞（）：；／％"
translate_table = str.maketrans(s1, s2)    #For converting between ASCII and fullwidth SHIFT-JIS characters

def decode_text(s):
    s1 = b''
    while '{' in s:
        s1 += s[:s.find('{')].translate(translate_table).encode('cp932')
        s1 += s[s.find('{') + 1:s.find('}')].encode('ascii')
        s = s[s.find('}') + 1:]
    s1 += s.translate(translate_table).encode('cp932')
    return s1

def load_input(filename):
    inputdata = []
    with open(filename, 'r', encoding='utf-8') as f:
        for line in f:
            line = line.rstrip('\r\n').split('\t')
            if line[0] != '':               #New entry
                ptrs = []
                for ptr in line[PTR_COL:]:  
                    if is_hex(ptr):
                        ptr = int(ptr, 16)
                    else:                   #End of pointer list
                        break
                    ptrs.append(ptr)
                inputdata.append([int(line[0], 16) - 0x20, line[INPUT_COL], ptrs]) #addr, text, ptrs
            else:                           #Additional line for same entry
                inputdata[-1][1] += '\x0a' + line[INPUT_COL]
    return inputdata

def is_hex(n):
    try:
        int(n, 16)
    except ValueError:
        return False
    return True

with open('ULJS00190.BIN', 'rb') as f:
    info = eboot.ELFinfo(f)
    filedata = bytearray(f.read())
#Add extra space for strings using eboot module
expansion_size = 0x6000
expansion_pos = 0x16C8B4
info.ELFh.shoff += expansion_size       
info.ph[0].filesz += expansion_size     #Normally you need to adjust memsz as well
filedata[expansion_pos:expansion_pos] = b'\xFF' * expansion_size
for sh1 in info.sh:
    if sh1.offset > expansion_pos:
        sh1.offset += expansion_size
    if sh1.offset + sh1.size == expansion_pos and sh1.size != 0:
        sh1.size += expansion_size
info.writeback(filedata)
##with open('EBOOT.BIN', 'rb') as f:
##    f.write(filedata)

#Insert SHIFT-JIS text (which is most of it)
inputdata = load_input('SJIS.tsv')
nextpos = 0x16C8B4 + 0x20                         #Start of extra data (expanded EBOOT)
for addr, text, ptrs in sorted(inputdata, key=lambda x: x[0], reverse=True):
##    if addr == 0x15ada8 - 0x20:
##        pdb.set_trace()
#Go backwards (reverse = True) // key is addr - we want to do later lines first
    endpos = filedata.find(b'\x00', addr)   #Find end of string
    while filedata[endpos] == 0:            #Find start of next string
        endpos += 1
    endpos -= 2                             #Need 2 nulls (at least) for SJIS terminator (I think)
    avail_space = endpos - addr             
    text = decode_text(text)
    if len(text) > avail_space:                         #NOT enough space
        filedata[addr:endpos] = b'\x00' * avail_space   #Blank out the line
        text += b'\x00' * 2                             #Add terminator
        while len(text) % 4 != 0:                       #Align to word boundary
            text += b'\x00'
        filedata[nextpos:nextpos+len(text)] = text      #Insert text
#Compute virtual address (vaddr) - 0x89C4200 is the difference between file
#image position and virtual address for text in the EBOOT extra data area
        vaddr = nextpos + 0x08b31010 - 0x16C8B4         #0x89C4200
        for ptr in ptrs:
            if ptr > 0x08000000:                        #Virtual address specified in input
#0x8803F8C is the difference between file image position and virtual address
#for text in the original EBOOT area (not the extra data section)
                vaddr_orig = addr + 0x8803FAC           
                if (vaddr & 0xFFFF) > 0x8000:           #Accomodate addiu instructions (all of them are addiu)
                    vaddr += 0x10000
                ptr = ptr - 0x8803FAC                   #Find file address of ptr
                filedata[ptr:ptr+2] = struct.pack('<H', vaddr >> 16)        #Write high bytes
                ptr = filedata.find(struct.pack('<H', vaddr_orig & 0xFFFF), 
                                    ptr)                                    #Search for original low bytes
                if ptr % 4 != 0:                                            #Error on search misses (was really for debugging)
                    print('{} Error: ptr={}, searchedfor={}'.format(
                        hex(addr), hex(ptr),  binascii.hexlify(struct.pack('<H', vaddr_orig & 0xFFFF))))
                filedata[ptr:ptr+2] = struct.pack('<H', vaddr & 0xFFFF)     #Write low bytes
            else:                                       #File address specified in input (this is a lot simpler)
                filedata[ptr:ptr+4] = struct.pack('<I', vaddr)              #Write the pointer
        nextpos += len(text)                            #Update state (position next text should go)
    else:                                               #THERE IS enough space
        text += b'\x00' * (avail_space - len(text))     #Pad the text out to the right length
        filedata[addr:endpos] = text                    #Replace the text
#Insert UTF-8 text
inputdata = load_input('UTF8.tsv')
for addr, text, ptrs in inputdata:
    endpos = filedata.find(b'\x00', addr)   #Find end of string
    avail_space = endpos - addr
    text = text.encode('UTF-8')
    text = text.replace(b'{', b'\\n')
    if len(text) > avail_space:
        print('{} too long. Text len={}, avail space={}'.format(hex(addr), len(text), avail_space))
    else:
        text = text + b'\x00' * (avail_space - len(text))
        filedata[addr:endpos] = text

#Build movie.bin and update TOC
FAT2 = movie_bin_tool.pack()
pos = 0x1628b0
filedata[pos:pos + 4 * 2 * 13] = \
    b''.join([struct.pack('<II', *x) for x in FAT2])

#Build shared.bin and update TOC
FAT = shared_bin_tool.pack()
pos = 0x1509B0
filedata[pos:pos+4] = struct.pack('<I', FAT[0][1])      #shared.bin LBA addr
pos = 0x15DC84
filedata[pos:pos + 4 * 3 * 24] = \
    b''.join([struct.pack('<III', *x) for x in FAT])

with open('EBOOT.BIN', 'wb') as f:
    f.write(filedata)

#Run armips
subprocess.run(['armips', 'mod.asm'])
